#!/usr/bin/perl -w
##
# DCPU-16 assembler I/O functions for reading and writing binary files.
#
# The specification doesn't say anything about the 8bit representation of
# a 16bit word output. The backstory only talks about the 16bit word
# ordering. Consequently, the representation could be either way around.
#
# By default we assume little endian, although an explicit ordering can
# be forced by using a different file extension:
#
# @li  `.bin`    - default ordering (little endian)
# @li  `.binle`  - little-endian
# @li  `.binbe`  - big-endian
#
# Binary dumps contain no symbols or relocation data. They are raw
# dumps of the content.
#
# @file
# @author Justin Fletcher
#

package DASMIO::Binary;

use DASM;

## Default endian order; use 'binle' or 'binbe'.
my $default_endian = 'binle';


##
# Create an object upon which we can work.
#
# @param[in] $proto  Prototype object, or this class's name
# @param[in] $dasm   DASM object we're working on
#
# @return new object
sub new
{
    my $proto = shift;
    my $dasm  = shift;
    my $class = ref($proto) || $proto;
    my $self  = {};

    bless $self, $class;
    
    $self->{'dasm'} = $dasm;
    
    return $self;
}


##
# Identify whether we can handle reading or writing a given file
# by its filename.
#
# @param[in]  $self     Address object
# @param[in]  $filename Filename we're accessing
# @param[in]  $type     'read' or 'write'
#
# @retval 1 if we can access it
# @retval 0 if we cannot access it
sub identify
{
    my ($self, $filename, $type) = @_;
    
    if ($filename =~ /\.bin(le|be)?$/)
    {
        return 1;
    }
    
    return 0;
}


##
# List the extensions we support
#
# @param[in]  $self     Address object
#
# @return arrayref of extensions
sub extensions
{
    return [ 'bin', 'binle', 'binbe' ];
}


##
# Read a binary dump and populate our core.
#
# @param[in]  $self     Address object
# @param[in]  $content  Content to parse
# @param[in]  $filename File it came from
sub read
{
    my ($self, $content, $filename) = @_;
    my $dasm = $self->{'dasm'};
    
    my ($format) = ($filename =~ /\.(\w+)$/);
    $format = $default_endian if ($format eq 'bin');

    my $packing;
    $packing = 'v*' if ($format eq 'binle');
    $packing = 'n*' if ($format eq 'binbe');
    
    for my $value (unpack $packing, $content)
    {
        $dasm->storetopc($value);
    }
}


##
# Write a binary dump from our core.
#
# In order to write an address dump, we need to force a resolve of all
# of the symbols. If there are any symbols that cannot be resolved, that's
# a fatal error - because they would write out as 0s.
#
# @param[in]  $self     Address object
# @param[in]  $filename Filename we're writing
#
# @return     Content to write
sub write
{
    my ($self, $filename) = @_;
    my $dasm = $self->{'dasm'};
    my $core = $dasm->{'core'};
    
    if ($dasm->resolve())
    {
        # everything was resolved
    }
    else
    {
        my %unresolved = reverse %{ $dasm->{'relocations'} };
        my $symbols = "Unresolved symbols:\n  " .
                      join "\n",
                        map {
                            "  $_"
                        } keys %unresolved;
        $dasm->error("Could not resolve all symbols to write out Address file\n$symbols");
    }
    
    # Determine the format we're using
    my ($format) = ($filename =~ /\.(\w+)$/);
    $format = $default_endian if ($format eq 'bin');
    
    my $packing;
    $packing = 'v*' if ($format eq 'binle');
    $packing = 'n*' if ($format eq 'binbe');
    
    my @content;
    my $lastaddress = -1;
    for my $address (sort { $a <=> $b } keys %$core)
    {
        my $value = $core->{$address};
        
        # We never write out hardware
        next if (ref $value ne '');

        while ($lastaddress+1 != $address)
        {
            push @content, 0;
        }
        
        push @content, $value;
        
        # Remember where we were
        $lastaddress = $address;
    }
    
    return pack($packing, @content);
}


# Return a true value
1;
