#!/usr/bin/perl -w
##
# DASM DCPU-16 source assembler class.
#
# This class provides the implementation of the source code assembler
# for DASM.
#
# The files we support end in '.dasm'.
#
# Source file format:
#
# @code
# [<label>]    <opcode>       <operand a>, <operand b>
# [<label>]    <ext-opcode>   <operand>
# ; <comment>
# @endcode
#
#
# It also provides a disassembler which can produce source code which
# can be passed back to the assembler. If the source comes from one
# of the binary formats that doesn't offer relocations or symblols,
# the content will be more difficult to parse interpret, but should still
# work.
# 
# The structures we use are intended to be able to support libraries of
# relocatable, partially linked code. This has not yet been tested and
# the lack of any indication of which symbols are local and which are
# imported may be a problem in the future.
#
# In addition to the 1.1 DCPU-16 specification, the assembler supports:
# 
# - Character literals for values: @n
#   `'X'` will evaluate to the number 88.
#
# - Expression calculations: @n
#   `1+4` will evaluate to the number 5. @n
#   Expression calculations are limited, and do not honour precedence
#   rules. Use brackets for clarity.
#
# - Symbols may be used in additive calculations: @n
#   `data + 4` will evaluate to the value of symbol '`data`' plus 4. @n
#   Symbols may (internally) be addresses or constants:
#   - Address symbols are created through the standard labels.
#   - Address symbols may (in future versions) be subject to relocations.
#
# - Symbols may be 'local' by preceding with a '`$`' character: @n
#   Any symbol preceded by a '`$`' will have the last label used prepended
#   to the symbol name. This allows labels and references to be used
#   within a routine without polluting the global namespace.
#
# - Additional instructions:
#   - `DAT <value[,<value> ]*` @n
#     ... will write literal values to the core. @n
#     The DAT instruction is supported in the assembler, and generated in
#     the disassembler for undefined instruction sequences.
#   - `CONST value` @n
#     ... will define a constant symbol with the name given in the label.
#   - `ADDRESS <value>` @n
#     ...  will define an address symbol with the name given in the label.
#
# - Directives: @n
#   `.<directive> <parameters>` allows additional meta operations to be
#   performed. Directives supported:
#   - `.ORIGIN address` @n
#     ... Sets the address for subsequent compilation
#   - `.INCLUDE file` @n
#     ...Includes another file at the current location
#
# @note We should really consider changing the structure so that the
#       DASMIO objects inherit from a parser class, rather than having
#       to reference `$self->{'dasm'}` for parser operations like
#       error().
#
# @file
# @author Justin Fletcher
#

package DASMIO::Source;

use DASM;
use DCPU;

use Carp;


## Whether we're debugging the 'DAT' instruction
my $debug_dat = 0;


## Whether we're debugging the symbol processing
my $debug_sym = 0;

##
# Create an object upon which we can work.
#
# @param[in] $proto  Prototype object, or this class's name
# @param[in] $dasm   DASM object we're working on
#
# @return new object
sub new
{
    my $proto = shift;
    my $dasm  = shift;
    my $class = ref($proto) || $proto;
    my $self  = {

            # The last label we encountered, for local labelling
            'lastlabel' => undef,
        };

    bless $self, $class;
    
    $self->{'dasm'} = $dasm;
    
    return $self;
}


##
# Identify whether we can handle reading or writing a given file
# by its filename.
#
# @param[in]  $self     Our object
# @param[in]  $filename Filename we're accessing
# @param[in]  $type     'read' or 'write'
#
# @retval 1 if we can access it
# @retval 0 if we cannot access it
sub identify
{
    my ($self, $filename, $type) = @_;
    
    if ($filename =~ /\.dasm$/)
    {
        return 1;
    }
    
    return 0;
}


##
# List the extensions we support.
#
# @param[in]  $self     Our object
#
# @return arrayref of extensions
sub extensions
{
    return [ 'dasm' ];
}


##
# Parse an instruction line.
#
# @param[in] $self    IO object
# @param[in] $line    Line to process
sub parseline
{
    my ($self, $line) = @_;
    my $dasm = $self->{'dasm'};
    
    # Trim any trailing newline that might have been left
    chomp $line;
    
    # Directives processed first - consider moving these to a callback.
    if ($line =~ s/^\.([A-Za-z]+)\s*//)
    {
        my $directive = uc $1;
        if ($directive eq 'ORIGIN')
        {
            my ($constant, $symname) = $self->parse_constant(\$line, 'constant');
            $dasm->{'pc'} = $constant;
        }
        elsif ($directive eq 'INCLUDE')
        {
            if ($line =~ s/"(.*?)"//)
            {
                $dasm->readfile($1);
            }
            else
            {
                $dasm->error("Unrecognised INCLUDE '$line'");
            }
        }
        else
        {
            $dasm->error("Unrecognised directive '$directive'");
        }
    }
    
    # We hold on to the label and PC location so that we can assign at the
    # end if they were set.
    my $label = undef;
    my $startpc = $dasm->{'pc'};

    # The initial part of the line may be a label, starting with a ':'
    # However, the label isn't always for this address; it may be a constant
    # label or some other type of label. So we merely remember it until we
    # know what sort of label it is.
    if ($line =~ s/^:([\w\$]+)//)
    {
        $label = $1;
        if ($label =~ /^\$/)
        {
            # A local label, so prepend the last label we saw
            if (!defined $self->{'lastlabel'})
            {
                $dasm->error("Cannot use local label without a recent label");
            }
            $label = $self->{'lastlabel'} . $label;
        }
        else
        {
            $self->{'lastlabel'} = $label;
        }
    }
    
    # Skip any indentation
    $line =~ s/^\s*//;
    
    # Skip any empty lines or comments
    if ($line eq '' ||
        $line =~ s/^;//)
    {
        goto line_complete;
    }
    
    # Extract the opcode, if there is one
    if ($line =~ s/^([a-zA-Z0-9_]+)\b\s*//)
    {
        my $opcode = uc $1;
        
        # First check for magic opcodes that we introduced:
        if ($opcode eq 'DAT')
        {
            # Parse operands until we have none left
            while ($line !~ /^\s*$/ &&
                   $line !~ /^\s*;/)
            {
                # We want to allow string data as well, for simplicity.
                # We do NOT handle UTF-8 - everything is 8 bit.
                if ($line =~ s/^\s*"([^"]+)"// || # "d string
                    $line =~ s/^\s*'([^']+)'//)   # 'd string
                {
                    print "  string: '$1'\n" if ($debug_dat);
                    for my $chr (unpack "c*", $1)
                    {
                        $dasm->storetopc($chr);
                    }
                }
                else
                {
                    my ($constant, $symname) = $self->parse_constant(\$line);
                    if ($symname)
                    {
                        print "  symbol $symname\n" if ($debug_dat);
                        # Relocation symbol, update the table
                        $dasm->{'relocations'}->{ $dasm->{'pc'} } = $symname;
                    }
                    print "  value $constant\n" if ($debug_dat);
                    $dasm->storetopc($constant);
                }
                last if ($line !~ s/^\s*,\s*//);
            }
            if ($line !~ s/^\s*// &&
                $line !~ s/^\s*;//)
            {
                $dasm->error("Garbage after operand to opcode '$opcode'");
            }
        }
        elsif ($opcode eq 'CONST')
        {
            # Constant value symbol declaration
            if (defined $label)
            {
                my ($constant, $symname) = $self->parse_constant(\$line, 'constant');
                $dasm->setsymbol($label, $constant, 'constant');
                $label = undef;
            }
            else
            {
                $self->error("CONST must be supplied with a label");
            }
        }
        elsif ($opcode eq 'ADDRESS')
        {
            # Address value symbol declaration
            if (defined $label)
            {
                my ($constant, $symname) = $self->parse_constant(\$line, 'constant');
                $dasm->setsymbol($label, $constant, 'address');
                $label = undef;
            }
            else
            {
                $self->error("ADDRESS must be supplied with a label");
            }
        }

        # Now the real opcodes
        elsif ($DCPU::BasicOps{$opcode})
        {
            # We have a basic opcode
            my $opval = $DCPU::BasicOps{$opcode};
            
            # Parse each operand
            my $operand_a = $self->parse_operand(\$line);
            if ($line !~ s/^\s*,\s*//)
            {
                $dasm->error("Expected operand b to opcode '$opcode'");
            }

            my $operand_b = $self->parse_operand(\$line);
            if ($line !~ s/^\s*$// &&
                $line !~ s/^\s*;//)
            {
                $dasm->error("Garbage after operand b to opcode '$opcode'");
            }

            if ($debug_asm)
            {
                printf "%04x : %-6s : %x\n", $dasm->{'pc'}, $opcode, $opval;
                printf "  %02x : %-9s : %s\n", $operand_a->{'opval'},
                                               $operand_a->{'type'},
                                               ($operand_a->{'value'} // '');
                printf "  %02x : %-9s : %s\n", $operand_b->{'opval'},
                                               $operand_b->{'type'},
                                               ($operand_b->{'value'} // '');
            }
            $dasm->storetopc($opval +
                             ($operand_a->{'opval'} << 4) +
                             ($operand_b->{'opval'} << 10));
            
            if (defined $operand_a->{'nextword'})
            {
                if ($operand_a->{'nextword'}->{'symbol'})
                {
                    # Relocation symbol, update the table
                    $dasm->{'relocations'}->{ $dasm->{'pc'} } = $operand_a->{'nextword'}->{'symbol'};
                }
                $dasm->storetopc($operand_a->{'nextword'}->{'value'});
            }

            if (defined $operand_b->{'nextword'})
            {
                if ($operand_b->{'nextword'}->{'symbol'})
                {
                    # Relocation symbol, update the table
                    $dasm->{'relocations'}->{ $dasm->{'pc'} } = $operand_b->{'nextword'}->{'symbol'};
                }
                $dasm->storetopc($operand_b->{'nextword'}->{'value'});
            }
        }
        elsif ($DCPU::ExtendedOps{$opcode})
        {
            # We have an extended opcode
            my $opval = $DCPU::ExtendedOps{$opcode};
            
            # Parse the operand
            my $operand_a = $self->parse_operand(\$line);
            if ($line !~ s/^\s*$// &&
                $line !~ s/^\s*;//)
            {
                $dasm->error("Garbage after operand a to opcode '$opcode'");
            }

            $dasm->storetopc(($opval<<4) +
                             ($operand_a->{'opval'}<<10));
            
            # Update the core with the following words
            if (defined $operand_a->{'nextword'})
            {
                if ($operand_a->{'nextword'}->{'symbol'})
                {
                    # Relocation symbol, update the table
                    $dasm->{'relocations'}->{ $dasm->{'pc'} } = $operand_a->{'nextword'}->{'symbol'};
                }
                $dasm->storetopc($operand_a->{'nextword'}->{'value'});
            }
        }
        else
        {
            $dasm->error("Bad opcode '$opcode'");
        }
    }
    else
    {
         $dasm->error("Bad source '$line'");
    }

line_complete:
    if (defined $label)
    {
        $dasm->label($startpc, $label);
    }
}


##
# Parse operands and return the values of them
#
# @param[in]     $self      IO object
# @param[in,out] $lineref   Reference to the line that we're parsing
#
# @return hashref of operand details:
#           'type'  =>  'reg'       (register value : register=>regname)
#                       'regaddr'   (register address : register=>regname)
#                       'regindex'  (registered indexed : register>regname)
#                       'pop'       (stack pop)
#                       'peek'      (stack peek)
#                       'push'      (stack push)
#                       'sp'        (stack pointer)
#                       'pc'        (program counter)
#                       'o'         (overflow)
#                       'indirect'  (dereference next address)
#                       'load'      (load next address)
#                       'loadshort' (load short value : value=>short value)
#           'opval' => value in the operand register
#           'register' => register, if applicable
#           'nextword' => next word data, if applicable
#           'value'    => value, if applicable
sub parse_operand
{
    my ($self, $lineref) = @_;
    my $dasm = $self->{'dasm'};
    
    #0x00-0x07: register (A, B, C, X, Y, Z, I or J, in that order)
    if ($$lineref =~ s/^([ABCXYZIJ])\b//i)
    {
        return {
                'type'     => 'reg',
                'register' => uc $1,
                'opval'    => 0 + $DCPU::RegisterMap{uc $1}
            };
    }
    #0x08-0x0f: [register]
    elsif ($$lineref =~ s/^\[\s*([ABCXYZIJ])\s*\]//i)
    {
        return {
                'type'     => 'regaddr',
                'register' => uc $1,
                'opval'    => 0x08 + $DCPU::RegisterMap{uc $1}
            };
    }
    #0x10-0x17: [next word + register]
    #     0x1e: [next word]
    elsif ($$lineref =~ s/^\[\s*//)
    {
        # Variation on syntax, as part of the tests at:
        # http://0x10cwiki.com/wiki/Comparison_of_Developer_Tools
        # allows for [register + next word] as well.
        if ($$lineref =~ s/^([ABCXYZIJ])\s*\+\s*//i)
        {
            my $reg = uc $1;
            # This is the variation.
            my $symbol = $self->parse_symbol($lineref);
            if ($$lineref =~ s/^\s*\]//)
            {
                return {
                        'type'     => 'regindex',
                        'register' => $reg,
                        'nextword' => $symbol,
                        'opval'    => 0x10 + $DCPU::RegisterMap{$reg}
                    };
                
            }
            else
            {
                $dasm->error("Garbage whilst parsing inverted indexed operand");
            }
        }
        else
        {
            my $symbol = $self->parse_symbol($lineref);
            if ($$lineref =~ s/^\s*\+\s*([ABCXYZIJ])\s*\]//i)
            {
                return {
                        'type'     => 'regindex',
                        'register' => uc $1,
                        'nextword' => $symbol,
                        'opval'    => 0x10 + $DCPU::RegisterMap{uc $1}
                    };
            }
            elsif ($$lineref =~ s/^\s*\]//)
            {
                return {
                        'type'     => 'indirect',
                        'nextword' => $symbol,
                        'opval'    => 0x1e
                    };
            }
            else
            {
                $dasm->error("Garbage whilst parsing referenced operand");
            }
        }
    }
    #     0x18: POP / [SP++]
    elsif ($$lineref =~ s/^POP\b//)
    {
        return {
                'type'     => 'pop',
                'opval'    => 0x18
            };
    }
    #     0x19: PEEK / [SP]
    elsif ($$lineref =~ s/^PEEK\b//)
    {
        return {
                'type'     => 'peek',
                'opval'    => 0x19
            };
    }
    #     0x1a: PUSH / [--SP]
    elsif ($$lineref =~ s/^PUSH\b//i)
    {
        return {
                'type'     => 'push',
                'opval'    => 0x1a
            };
    }
    #     0x1b: SP
    elsif ($$lineref =~ s/^SP\b//i)
    {
        return {
                'type'     => 'sp',
                'opval'    => 0x1b
            };
    }
    #     0x1c: PC
    elsif ($$lineref =~ s/^PC\b//i)
    {
        return {
                'type'     => 'pc',
                'opval'    => 0x1c
            };
    }
    #     0x1d: O
    elsif ($$lineref =~ s/^O\b//)
    {
        return {
                'type'     => 'o',
                'opval'    => 0x1d
            };
    }
    #     0x1f: next word (literal)
    #0x20-0x3f: literal value 0x00-0x1f (literal)
    else 
    {
        # Literal value, which could be a short-form literal value,
        # or a symbol
        my $symbol = $self->parse_symbol($lineref);
        if ($symbol->{'type'} eq 'constant')
        {
            if ($symbol->{'value'} >= 0 &&
                $symbol->{'value'} < 0x20)
            {
                return {
                        'type'  => 'loadshort',
                        'opval' => 0x20 + $symbol->{'value'},
                        'value' => $symbol->{'value'}
                    };
            }
        }
        
        # Check for overflow
        if ($symbol->{'type'} eq 'constant' &&
            $symbol->{'value'} > 0xffff)
        {
            $dasm->error( sprintf "Literal value (0x%x) too large", $symbol->{'value'} );
        }
        # Check for underflow
        if ($symbol->{'type'} eq 'constant' &&
            $symbol->{'value'} < 0)
        {
            # I think we'll allow negative numbers to wrap down, as that's
            # useful in the assembler. But those which wrap twice are too
            # small.
            if ($symbol->{'value'} < -0x10000)
            {
                $dasm->error( sprintf "Literal value (0x%x) too small", $symbol->{'value'} );
            }
            $symbol->{'value'} += 0x10000;
        }
        return {
                'type'  => 'load',
                'opval' => 0x1f,
                'nextword' => $symbol
            };
    }
}


##
# Parse a symbol name or value.
#
# @param[in]     $self      DASM object
# @param[in,out] $lineref   Reference to the line that we're parsing
#
# @return hashref of symbol details:
#           'type'  =>  'constant' : a constant value
#                       'address'  : an address constant
#           'value' =>  value
#           'symbol' => any symbol name to use as relocation
sub parse_symbol
{
    my ($self, $lineref, $nonterm) = @_;
    my $dasm = $self->{'dasm'};

    my %symbol = (
            'type'  => 'unknown',
            'value' => 0,
        );
    
    if (!defined $lineref)
    {
        confess "parse_symbol called without a parameter";
    }
    if (ref $lineref ne 'SCALAR')
    {
        confess "parse_symbol called with ".ref($lineref)." but expected SCALAR";
    }

    # Skip any leading spaces
    $$lineref =~ s/^\s*//;
    
    print "parse_symbol: $$lineref\n" if ($debug_sym);
    
    if ($$lineref =~ s/^\(//)
    {
        # Opening bracket for an expression, so parse inside the
        # brackets
        my $innersymbol = $self->parse_symbol($lineref);
        
        %symbol = %$innersymbol;
        
        # Check that we ended properly
        if ($$lineref !~ s/^\s*\)//)
        {
            $dasm->error("Missing closing bracket");
        }
    }
    elsif ($$lineref =~ s/^([\$A-Za-z_][\w\$]*)//)
    {
        # Symbol reference
        my $symname = $1;

        # Check for local symbols and redefine them for the local label
        if ($symname =~ /^\$/)
        {
            if (!defined $self->{'lastlabel'})
            {
                $dasm->error("Cannot use a local label before a label has been set");
            }
            $symname = $self->{'lastlabel'} . $symname;
        }

        my $sym = $dasm->getsymbol($symname);
        # Hmm. Whilst decoding the symbol here makes some sense, we could
        # just leave it for the linkage stage.
        #if (defined $sym)
        #{
        #    # Symbol has been defined
        #    %symbol = ( %$sym,
        #                'symbol' => $symname
        #              );
        #}
        #else
        {
            %symbol = (
                    'type'   => 'address',
                    'value'  => 0,
                    'symbol' => $symname
                );
        }
    }
    elsif ($$lineref =~ s/^([+-]?0x[A-Fa-f0-9]+)// ||
           $$lineref =~ s/^([+-]?\d+)//)
    {
        # Hex or decimal value
        my $value = eval $1;
        %symbol = (
                'type' => 'constant',
                'value' => $value
            );
    }
    elsif ($$lineref =~ s/^'(.)'//)
    {
        # character, so convert to an ordinal
        %symbol = (
                'type' => 'constant',
                'value' => ord($1)
            );
    }
    else
    {
        $dasm->error("Unrecognised symbol parsing '$$lineref'");
    }
    
    # We need to special case the syntax here, because the defined format
    # of the assembler allows for register names here (this could be
    # refined by having a more complex symbol definition).
    if ($$lineref =~ /\+\s*[ABCXYZIJ]\b/i)
    {
        # This is a relative reference, so don't try to perform arithmetic
        # here.
        return \%symbol;
    }
    
    if ($$lineref =~ s/^\s*([\+\-\*\/\%\&\|\^]|<<|>>)//)
    {
        # Operators - note that this is wrong WRT precedence as it won't
        # make *, / and % higher priority than + and -. I'm not fussed
        # right now. FIXME
        my $operator = $1;
        my $second = $self->parse_symbol($lineref);
        if (defined $symbol{'symbol'} &&
            $operator ne '+')
        {
            $dasm->error("Symbol '$symbol{'symbol'}' not yet defined in arithmetic operation (only additive operations allowed)");
        }
        if (defined $symbol{'symbol'} &&
            defined $second->{'symbol'})
        {
            $dasm->error("Relocation by multiple symbols ('$symbol{'symbol'}' and '$second->{'symbol'}') not supported");
        }
        
        if (defined $second->{'symbol'})
        {
            # Swap them around so that the symbol is in our local
            # This is safe because the only type of symbol operation we
            # support is additive.
            my %copy = %symbol;
            %symbol = %$second;
            $second = \%copy;
        }
        
        if ($operator eq '+')
        { $symbol{'value'} += $second->{'value'}; }
        elsif ($operator eq '-')
        { $symbol{'value'} -= $second->{'value'}; }
        elsif ($operator eq '*')
        { $symbol{'value'} *= $second->{'value'}; }
        elsif ($operator eq '/')
        { $symbol{'value'} /= $second->{'value'}; }
        elsif ($operator eq '%')
        { $symbol{'value'} %= $second->{'value'}; }
        elsif ($operator eq '&')
        { $symbol{'value'} &= $second->{'value'}; }
        elsif ($operator eq '|')
        { $symbol{'value'} |= $second->{'value'}; }
        elsif ($operator eq '^')
        { $symbol{'value'} ^= $second->{'value'}; }
        elsif ($operator eq '<<')
        { $symbol{'value'} <<= $second->{'value'}; }
        elsif ($operator eq '>>')
        { $symbol{'value'} >>= $second->{'value'}; }
        else
        {
            die "Rogue operator $operator - should not happen";
        }

        if ($second->{'type'} eq 'address')
        {
            $symbol{'type'} = $second->{'type'};
        }
    }
    
    return \%symbol;
}

##
# Obtain a constant from an input line.
#
# @param[in]     $self      DASMIO object
# @param[in,out] $lineref   reference to line to read from
# @param[in]     $symbol    'constant' : only a constant is acceptable
#                           'symbol'   : symbol is acceptable
#
# @return $value    The value to store
#         $symbol   Relocation symbol to apply, or none if not found
sub parse_constant
{
    my ($self, $lineref, $symbol) = @_;
    
    $symbol = 'constant' if (!defined $symbol);
    
    print "Constant: $$lineref\n" if ($debug_dat);

    my $operand = $self->parse_operand($lineref);
    print "  type $operand->{'type'}\n" if ($debug_dat);

    # The only types of operand we support here are the
    # nextword and constant value operands - 'load' and 'loadshort'
    my $sym = undef;
    if ($operand->{'type'} eq 'load')
    {
        $sym = $operand->{'nextword'};
    }
    elsif ($operand->{'type'} eq 'loadshort')
    {
        $sym = {
                'type' => 'constant',
                'value' => $operand->{'value'}
            };
    }
    else
    {
        $dasm->error("Constant operand expected, not type $operand->{'type'}");
    }

    if ($sym->{'symbol'} &&
        $symbol eq 'constant')
    {
        # A symbol which we know about and which is a constant.
        my $symval = $dasm->getsymbol($symname);
        if (defined $symval &&
            $symval->{'type'} eq 'constant')
        {
            # We can interpret it here immediately
            return ($sym->{'value'} + $symval->{'value'}, undef);
        }
        else
        {
            $dasm->error("Symbol operand not expected");
        }
    }
    
    return ($sym->{'value'}, $symval->{'symbol'});
}



###########################################################################
# Disassembly routines
# (these are included in the source parser, so that we can decode the
# compiled files to re-compilable files)


##
# Write out a compilable source file based on what we have in the core.
#
# @param[in]  $self     Address object
#
# @return     Content to write
sub write
{
    my ($self) = @_;
    my $dasm = $self->{'dasm'};
    my $core = $dasm->{'core'};

    # Easier access to symbols
    my $symbols = $dasm->{'symbols'};
    
    # Collect a list of address symbols
    my %addressSymbols = map {
            $symbols->{$_}->{'value'} => $_
        } grep {
            $symbols->{$_}->{'type'} eq 'address'
        } keys %$symbols;
    
    # The lines we'll return
    my @lines;
    
    # If we have any constants, we need to write them out first
    my @constants = grep {
                        $symbols->{$_}->{'type'} eq 'constant'
                    } keys %$symbols;
    if (scalar(@constants) != 0)
    {
        push @lines, "; Constants";
    }
    for my $sym (@constants)
    {
        push @lines, sprintf "%-16s%-8s0x%04x",
                             ":$sym ",
                             "CONST",
                             $symbols->{$sym}->{'value'};
    }

    my $lastaddress = -1;
    
    # We have to deal with the addresses in order, because we might
    # have a corrupt file that ends before the data or which is
    # discontinuous.
    my @addresses = sort { $a <=> $b } keys %$core;
    if (scalar(@addresses) != 0)
    {
        push @lines, "; Core";
    }
    while (defined (my $address = shift @addresses))
    {
        my $value = $core->{$address};
        
        # We never write out hardware
        next if (ref $value ne '');

        if ($lastaddress+1 != $address)
        {
            # If we have discontinous core data, we need to change the
            # origin.
            push @lines, sprintf ".ORIGIN 0x%04x", $address;
        }
        
        if (defined $addressSymbols{$address})
        {
            push @lines, ":$addressSymbols{$address}";
        }
        
        my $opcode = $value & 15;
        my $opcodename;
        
        # List of the memory locations associated with this instruction
        my @memdata = ($value);
        
        # List of the decoded operands
        my @operands; 
        if ($opcode !=0)
        {
            # Basic opcode
            $opcodename = $DCPU::InvBasicOps{$opcode};
            my $operand_a = ($value>>4) & 63;
            my $operand_b = ($value>>10) & 63;
            
            # Decode each operand into printable descriptions, pulling data
            # out of nextword as necessary.
            push @operands, $self->decode_operand($operand_a,
                                                  $address + scalar(@memdata),
                                                  \@memdata);

            push @operands, $self->decode_operand($operand_b,
                                                  $address + scalar(@memdata),
                                                  \@memdata);
            
            if (!defined $opcodename)
            {
                die "No such opcode: $opcode\n";
            }
        }
        else
        {
            # Extended opcodes
            $opcode = ($value>>4) & 63;
            $opcodename = $DCPU::InvExtendedOps{$opcode};
            if (!defined $opcodename)
            {
                # Invalid opcode, so we treat as data
                $opcodename = "DAT";
                @operands = ($value);
            }
            else
            {
                # a valid extended opcode
                my $operand_a = ($value>>10) & 63;

                push @operands, $self->decode_operand($operand_a,
                                                      $address + 1,
                                                      \@memdata);
            }
        }

        # Convert the memdata into presentable hex values
        @memdata = map { defined $_ ? sprintf "%04x", $_ : '----' } @memdata;

        # Construct the output line
        push @lines, sprintf "%-16s%-8s%-32s ; %04x: %s",
                                 "",
                                 $opcodename,
                                 (join ", ", @operands),
                                 $address,
                                 (join " ", @memdata);
        
        # Skip the data we've just read off the address list
        for (2..scalar(@memdata))
        {
            # We want to skip only the entries from the list that are actually
            # valid - the core might be discontinuous.
            if (defined $addresses[0] &&
                $addresses[0] == $address+1)
            {
                shift @addresses;
            }
            $address++;
        }
        $lastaddress = $address;
    }

    return join "\n", @lines, "";
}


##
# Decode an operand into a string.
#
# @param[in] $self     IO object
# @param[in] $operand  Operand value
# @param[in] $address  Address of 'nextword'
# @param[in,out] $memdataref    Reference to an address of data read from memory
#                               which we will update if we read any data
#
# @return Operand definition string
sub decode_operand
{
    my ($self, $operand, $address, $memdataref) = @_;
    
    # 0x00-0x07: register (A, B, C, X, Y, Z, I or J, in that order)
    if ($operand >= 0 && $operand <= 7)
    {
        return $DCPU::InvRegisterMap{ $operand & 7 };
    }
    # 0x08-0x0f: [register]
    elsif ($operand >= 0x8 && $operand <= 0xf)
    {
        return '[' . $DCPU::InvRegisterMap{ $operand & 7 } . ']';
    }
    # 0x10-0x17: [next word + register]
    elsif ($operand >= 0x10 && $operand <= 0x17)
    {
        my $nextword = $self->decode_nextword($address, $memdataref);
        return "[$nextword + " . $DCPU::InvRegisterMap{ $operand & 7 } . ']';
    }
    #      0x18: POP / [SP++]
    elsif ($operand == 0x18)
    {
        return 'POP';
    }
    #      0x19: PEEK / [SP]
    elsif ($operand == 0x19)
    {
        return 'PEEK';
    }
    #      0x1a: PUSH / [--SP]
    elsif ($operand == 0x1a)
    {
        return 'PUSH';
    }
    #      0x1b: SP
    elsif ($operand == 0x1b)
    {
        return 'SP';
    }
    #      0x1c: PC
    elsif ($operand == 0x1c)
    {
        return 'PC';
    }
    #      0x1d: O
    elsif ($operand == 0x1d)
    {
        return 'O';
    }
    #      0x1e: [next word]
    elsif ($operand == 0x1e)
    {
        my $nextword = $self->decode_nextword($address, $memdataref);
        return "[$nextword]";
    }
    #      0x1f: next word (literal)
    elsif ($operand == 0x1f)
    {
        my $nextword = $self->decode_nextword($address, $memdataref);
        return "$nextword";
    }
    #  0x20-0x3f: literal value 0x00-0x1f (literal)
    elsif ($operand >= 0x20 && $operand <= 0x3f)
    {
        return $operand - 0x20;
    }
    else
    {
        return "!Invalid: $operand!";
    }
}


##
# Read the next word from a memory address, whilst decoding operands.
# Apply any relocations that we know about such that the decode comes
# out as readable.
#
# The actual value in memory is appended to the memdataref list.
# If the memory is not known, we return '?' and an undef entry in the array.
#
# @param[in] $self     IO object
# @param[in] $address  Address of 'nextword'
# @param[in,out] $memdataref    Reference to an address of data read from memory
#                               which we will update if we read any data
#
# @return extword string
sub decode_nextword
{
    my ($self, $address, $memdataref) = @_;
    my $dasm = $self->{'dasm'};
    my $core = $dasm->{'core'};
    
    my $value = $dasm->{'core'}->{$address};

    # We're using it, so push it on to our stack
    push @$memdataref, $value;

    if (!defined $value)
    {
        # No value, which means that the core is truncated or discontinuous
        # around an invalid location, so we return '?'
        return '?';
    }

    # Check whether we have a relocation at that address
    if (defined $dasm->{'relocations'}->{$address})
    {
        # We do... which means that we have to perform report the symbol
        # of the relocation as part of a calculation.
        
        if ($value == 0)
        {
            # However, the offset appears to be 0, so we can just use the
            # symbol name.
            return "$dasm->{'relocations'}->{$address}";
        }
        else
        {
            return "($value + $dasm->{'relocations'}->{$address})";
        }
    }
    
    if ($value >= 256)
    {
        return sprintf "0x%04x", $value;
    }
    
    return sprintf "0x%02x", $value;
}


# Must return true
1;
