#!/usr/bin/perl -w
##
# Basic text emulation for the DASM assembler.
# This module provides the video RAM for a simple text array chip.
#
# The chip is 40x25 = 1000 bytes, out of the 1024 that are mapped.
# For simplicity, we use only the bottom 8 bits of the 16 bits for each
# character.
#
# @file
# @author Justin Fletcher
#

package DExecHW::BasicVideo;

use Curses;


##
# Construct a new hardware device object.
#
# @param[in]  $proto   This class object, or prototype to add to
# @param[in]  $object  The object to proxy for
# @param[in]  $offset  The offset this object is created for
#
# @return Assembler object
sub new
{
    my $proto = shift;
    my $object = shift;
    my $offset = shift || 0;
    my $class = ref($proto) || $proto;
    my $self = {
            'memory' => [
                    map { 0 } (0..0x3FF)
                ],
        };
    
    bless $self, $class;
    
    
    return $self;
}


##
# Start up the hardware.
#
# We perform all output through the Curses library.
#
# @param[in] $self   Object
sub start
{
    my ($self) = @_;

    Curses::initscr();

    $self->{'win'} = Curses::newwin(25, 40, 0, 0);
    #Curses::box( $self->{'win'}, 0,0);

    Curses::refresh( $self->{'win'} );
}


##
# Stop the hardware.
#
# @param[in] $self   Object
sub stop
{
    my ($self) = @_;

    Curses::endwin();
}



##
# Return the size of of the implementation.
#
# @param[in] $self   Object
#
# @return Size of the hardware implementation in words
sub window
{
    my ($self) = @_;

    return 0x400;
}


##
# Read value
#
# @param[in] $self   Object
# @param[in] $offset Offset of register within window
# @param[in] $exec   DExec object opon which this object is operating
#
# @return value of this register
sub read
{
    my ($self, $offset, $exec) = @_;

    $value = $self->{'memory'}->[$offset];
    
    return $value;
}


##
# Write value
#
# @param[in] $self   Object
# @param[in] $offset Offset of register within window
# @param[in] $exec   DExec object upon which this object is operating
# @param[in] $value  Value to write
#
# @return ignored
sub write
{
    my ($self, $offset, $exec, $value) = @_;

    $self->{'memory'}->[$offset] = $value;
    
    my $y = int($offset / 40);
    my $x = int($offset % 40);
    
    Curses::addch($self->{'win'}, $y, $x, $value);
    
    Curses::refresh( $self->{'win'} );
    
    #print "Write $value to $x,$y\n";
}


##
# Get the symbol name for an offset
#
# @param[in] $self   Object
# @param[in] $offset Offset of register within window
#
# @return symbol name to use
# @retval undef for a default symbol
# @retval "" for no symbol
sub symbol
{
    my ($self, $offset) = @_;
    
    return "";
}

# Must return 1
1;
