#!/usr/bin/perl -w
##
# HW implementation for 'SimpleKey'.
#
# A very simple key input.
#
# 2 words are provided (little endian):
#
# | Offset | Name  | Meaning |
# | ------ | ----- | ------- |
# | +0     | Read  | read a key, or 0 if none available |
# | +1     | Count | number of characters available |
#
# The characters available are dequeued by reading the Read register.
# Every read of the Read register will queue any new characters for
# reading.
#
# Writing to any register has no effect.
#
# @file
# @author Justin Fletcher
#

package DExecHW::SimpleKey;

use Term::ReadKey;


##
# Construct a new hardware device object.
#
# @param[in]  $proto   This class object, or prototype to add to
# @param[in]  $object  The object to proxy for
# @param[in]  $offset  The offset this object is created for
#
# @return Assembler object
sub new
{
    my $proto = shift;
    my $object = shift;
    my $offset = shift || 0;
    my $class = ref($proto) || $proto;
    my $self = {
            'queue' => [],
        };
    
    bless $self, $class;
    
    
    return $self;
}


##
# Start up the hardware.
#
# We configure the terminal for 'cbreak' mode which ensures that the
# keys are not buffered before delivery.
#
# @param[in] $self   Object
sub start
{
    my ($self) = @_;

    # Configure cbreak read mode, which delivers keys immediately
    ReadMode 'cbreak';
}


##
# Stop the hardware.
#
# Returns the terminal to the mode iti was in prior to being placed in
# 'cbreak' mode.
#
# @param[in] $self   Object
sub stop
{
    my ($self) = @_;

    # Configure cbreak read mode, which delivers keys immediately
    ReadMode 'restore';
}



##
# Return the size of of the implementation.
#
# @param[in] $self   Object
#
# @return Size of the hardware implementation in words
sub window
{
    my ($self) = @_;

    return 2;
}


##
# Read value
#
# @param[in] $self   Object
# @param[in] $offset Offset of register within window
# @param[in] $exec   DExec object opon which this object is operating
#
# @return value of this register
sub read
{
    my ($self, $offset, $exec) = @_;

    # Add any new keys to the queue
    while (my $key = ReadKey(-1))
    {
        push @{ $self->{'queue'} }, $key;
    }
    

    if ($offset == 1)
    {
        # Deal with the Count register.
        return scalar(@{ $self->{'queue'} });
    }
    
    # Return a value from the queue
    $key = shift @{ $self->{'queue'} };
    
    if (!defined $key)
    {
        $key = 0;
    }
    else
    {
        $key = ord($key);
    }
    
    return $key;
}


##
# Write value
#
# @param[in] $self   Object
# @param[in] $offset Offset of register within window
# @param[in] $exec   DExec object upon which this object is operating
# @param[in] $value  Value to write
#
# @return ignored
sub write
{
    my ($self, $offset, $exec, $value) = @_;
    # Not required
}


##
# Get the symbol name for an offset
#
# @param[in] $self   Object
# @param[in] $offset Offset of register within window
#
# @return symbol name to use
# @retval undef for a default symbol
# @retval "" for no symbol
sub symbol
{
    my ($self, $offset) = @_;
    
    return ("Read", "Count")[$offset];
}

# Must return 1
1;
