#!/usr/bin/perl -w
##
# Hardware execution adapter.
#
# This class provides a simple interface to hardware implementations.
# Instances of the class can be created which associate a hardware
# offset with the object instance. The instance will dispatch 'read' and
# 'write' functions with the offset.
#
# Example use:
#
# @code
#   my $exec; # Execution object
#   my @core;
#   for my $offset (0..15)
#   {
#       $core[$offset] = new DExecHW($hw, $offset);
#   }
#   $core[4]->write($number, $exec);    # will call $hw->write(4, $number, $exec);
#   my $value = $core[4]->read($exec);  # will call $hw->read(4, $exec);
# @endcode
#
# @file
# @author Justin Fletcher
#

package DExecHWAdapter;


##
# Construct a new hardware adapter object.
#
# @param[in]  $proto   This class object, or prototype to add to
# @param[in]  $object  The object to proxy for
# @param[in]  $offset  The offset this object is created for
#
# @return Assembler object
sub new
{
    my $proto = shift;
    my $object = shift;
    my $offset = shift || 0;
    my $class = ref($proto) || $proto;
    my $self = {
            'object' => $object,
            'offset' => $offset
        };

    bless $self, $class;
    
    return $self;
}


##
# Describe this object.
#
# @param[in]  $self   This Adapter object
#
# @return string describing this object
sub describe
{
    my ($self) = @_;
    return ref($self->{'object'}) . sprintf "[0x%x]", $self->{'offset'};
}


##
# Retrieve a symbol name for this object.
#
# @param[in]  $self   This Adapter object
#
# @return a symbol name for this register
# @retval undef if no symbol name should be assigned
sub symbol
{
    my ($self) = @_;

    if ($self->{'object'}->can('symbol'))
    {
        my $name = $self->{'object'}->symbol($self->{'offset'});
        return undef if (defined $name && $name eq '');
        return $name if (defined $name);
    }

    my $window = $self->{'object'}->window();
    if ($window >= 0x1000)
    {
        return sprintf("%04x", $self->{'offset'});
    }
    if ($window >= 0x100)
    {
        return sprintf("%03x", $self->{'offset'});
    }
    return sprintf("%02x", $self->{'offset'});
}


##
# 'Read' function; intended to be used to read data from a hardware
# register, supplying an offset.
#
# @param[in] $self   DExecHW object
# @param[in] @args   Arguments to pass to the object's read function,
#                    usually the execution environment object.
#
# @return read value
sub read
{
    my $self = shift;
    return $self->{'object'}->read($self->{'offset'}, @_);
}


##
# 'Write' function; intended to be used to write data to a hardware
# register, supplying an offset.
#
# @param[in] $self   DExecHW object
# @param[in] @args   Arguments to pass to the object's write function,
#                    usually the value to write and the execution
#                    environment object
#
# @return return value
sub write
{
    my $self = shift;
    return $self->{'object'}->write($self->{'offset'}, @_);
}


# Must return true
1;
