#!/usr/bin/perl -w
##
# Run the comparison suite from:
#   http://0x10cwiki.com/wiki/Comparison_of_Developer_Tools
# through the compiler to check that we conform well.
#
# The initial scores for DASM were:
#
# Group                    : Pass : Fail : Total
# ------------------------ : ---- : ---- : ----
# Errors                   :   17 :   13 :   30
# Instructions             :    1 :    0 :    1
# Notch                    :    0 :    1 :    1
# Operands                 :    0 :    1 :    1
# Test Code                :    1 :    0 :    1
# _TOTAL_                  :   19 :   15 :   34
#
# Failures are caused by:
#  * The tests assume that the order of the offset and register in the
#    indexed operand format is reversible. The specification states that
#    the format is 'nextword + register', not the converse.
#    Action: Update assembler to allow the 'register + nextword' format.
#
#  * errors-15 uses 'DIV a,0' as a test. This is an INVALID test as the
#    specification explicitly states that this is acceptable and results
#    in both a and O being set to 0. This may be a reasonable operation,
#    and produces a word in memory of 0x8002 which may be used for some
#    magic code.
#    Action: None. Test is invalid, and has been marked as such in this
#    test so that it not fail the regressions.
#
#  * errors-12 allows 'jsr a,a', when it should not be accepted.
#    errors-11 also fails for the same reason.
#    Action: Fix existing code; had a missing '^' anchor when checking
#    the following line content. A few places in the instruction and symbol
#    processing had omitted anchors; these were updated to correct the
#    checks.
#
#  * errors-3 checks for too large a constant, which is being accepted.
#    Action: Update the checks in the operand processing for constant
#    values being outside the range. Allow -ve values for -0x10000 to -1,
#    though, because they are useful.
#
#  * Test code was incorrectly penalising the first 'error' output for
#    the output generated by the previous test.
#    Action: Fixed the test code to look at the right file.
#
# Improvements based on results:
#  * 'Parse error: Could not resolve all symbols to write out Address file'
#    is produced if there are undefined symbols.
#    Action: The report now includes the unresolved symbols.
#


## The directory we execute within
my $testdir = "test-comparison";

## Test output file
my $testoutput = 'Xoutput.dump';

## A definition of the tests, containing the groups of tests we run,
# and the test files to assemble within those groups.
my %tests = (

    'Notch'         => [ 'notch.dasm' ],
    'Test Code'     => [ 'testcode.dasm' ],
    'Instructions'  => [ 'instructions.dasm' ],
    'Operands'      => [ 'operands.dasm' ],
    'Errors'        => [
        'errors-01.dasm',
        'errors-02.dasm',
        'errors-03.dasm',
        'errors-04.dasm',
        'errors-05.dasm',
        'errors-06.dasm',
        'errors-07.dasm',
        'errors-08.dasm',
        'errors-09.dasm',
        'errors-10.dasm',
        'errors-11.dasm',
        'errors-12.dasm',
        'errors-13.dasm',
        'errors-14.dasm',
        'errors-15.dasm',
    ],
);

# Invalid tests
my %invalid = (
    'errors-15.dasm' => 1
);

##
# We keep the scores for each level
my %scores = (
    '_TOTAL_' => {
            'pass'  => 0,
            'fail'  => 0,
            'total' => 0,
        }
);

# Go through the groups, running the tests.
for my $group (sort keys %tests)
{
    my $npass = 0;
    my $nfail = 0;
    
    print "Group '$group':\n";
    
    for my $test (@{ $tests{$group} })
    {
        my $cmd = "./dasm.pl -o $testoutput -i $testdir/$test";

        # Remove the output dump, which we don't care about.
        unlink $testoutput;
        
        my $output = `$cmd 2>&1`;
        my ($sig, $core, $rc) = (($? & 127), ($? & 128), ($? >> 8));
        my $failed;
        
        # Only the errors group is /expected/ to fail
        my $expectfail = 0;
        $expectfail = 1 if ($group =~ /Errors/);
        
        # Either a signal or core is *always* a failure
        if ($sig)
        {
            $failed = "Signal $sig";
        }
        elsif ($core)
        {
            $failed = 'Core dumped';
        }
        elsif ( ($rc != 0 && !$expectfail) ||
                ($rc == 0 && $expectfail) )
        {
            $failed = "Failed (rc=$rc)";
        }
        else
        {
            $failed = 'Passed';            
        }
        
        printf "  Test %-24s: $failed\n", "'$test'";
        
        # Check whether this test is even valid
        if (defined $invalid{$test})
        {
            print "    Invalid test - ignoring\n";
            next;
        }
        
        
        if ($group =~ /Errors/)
        {
            # We score an extra point for reporting the error
            if ($output ne '')
            {
                # If the output is empty then it's good.
                if (-f $testoutput)
                {
                    print "    (generated an output dump - should not have)\n";
                    unlink $testoutput;
                    $nfail++;
                }
                else
                {
                    $npass++;
                }
            }
            else
            {
                print "    No error report produced\n";
                $nfail++;
            }
        }
        if ($failed eq 'Passed')
        {
            $npass++;
        }
        else
        {
            $nfail++;
        }
        print map { "      $_\n" } split /\n/, $output;
    }
    
    $scores{$group} = {
        'pass'  => $npass,
        'fail'  => $nfail,
        'total' => $npass+$nfail,
    };
    $scores{'_TOTAL_'}->{'pass'} += $npass;
    $scores{'_TOTAL_'}->{'fail'} += $nfail;
    $scores{'_TOTAL_'}->{'total'} += $npass+$nfail;
}


# Print out the final scores
print "\n\n";
printf "%-24s : %4s : %4s : %4s\n",
       "Group",
       "Pass",
       "Fail",
       "Total";
printf "%-24s : %4s : %4s : %4s\n",
       '-' x 24,
       '-' x 4,
       '-' x 4,
       '-' x 4;
for my $group (sort keys %scores)
{
    printf "%-24s : %4i : %4i : %4i\n",
           $group,
           $scores{$group}->{'pass'},
           $scores{$group}->{'fail'},
           $scores{$group}->{'total'};
}

exit ( $scores{'_TOTAL_'}->{'fail'} != 0 );
